@@ -45,7 +45,6 @@
 #include "getopt.h"
 #endif				/* HAVE_GETOPT_H */
 
-#include "gtksecentry.h"
 #include "pinentry.h"
 
 #ifdef FALLBACK_CURSES
@@ -69,9 +68,12 @@ static GtkWidget *entry;
 static GtkWidget *repeat_entry;
 static GtkWidget *error_label;
 static GtkWidget *qualitybar;
+#if !GTK_CHECK_VERSION (2, 12, 0)
 static GtkTooltips *tooltips;
+#endif
 static gboolean got_input;
 static guint timeout_source;
+static int confirm_mode;
 
 /* Gnome hig small and large space in pixels.  */
 #define HIG_SMALL      6
@@ -93,6 +95,8 @@ constrain_size (GtkWidget *win, GtkRequisition *req, gpointer data)
   static gint width, height;
   GdkGeometry geo;
 
+  (void)data;
+
   if (req->width == width && req->height == height)
     return;
   width = req->width;
@@ -126,13 +130,16 @@ make_transient (GtkWidget *win, GdkEvent *event, gpointer data)
   GdkScreen *screen;
   GdkWindow *root;
 
+  (void)event;
+  (void)data;
+
   if (! pinentry->grab)
     return;
 
   /* Make window transient for the root window.  */
   screen = gdk_screen_get_default ();
   root = gdk_screen_get_root_window (screen);
-  gdk_window_set_transient_for (win->window, root);
+  gdk_window_set_transient_for (gtk_widget_get_window (win), root);
 }
 
 
@@ -140,10 +147,13 @@ make_transient (GtkWidget *win, GdkEvent *event, gpointer data)
 static int
 grab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
 {
+  (void)data;
+
   if (! pinentry->grab)
     return FALSE;
 
-  if (gdk_keyboard_grab (win->window, FALSE, gdk_event_get_time (event)))
+  if (gdk_keyboard_grab (gtk_widget_get_window (win),
+			 FALSE, gdk_event_get_time (event)))
     {
       g_critical ("could not grab keyboard");
       grab_failed = 1;
@@ -157,12 +167,14 @@ grab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
 static int
 ungrab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
 {
+  (void)data;
+
   gdk_keyboard_ungrab (gdk_event_get_time (event));
   /* Unmake window transient for the root window.  */
   /* gdk_window_set_transient_for cannot be used with parent = NULL to
      unset transient hint (unlike gtk_ version which can).  Replacement
      code is taken from gtk_window_transient_parent_unrealized.  */
-  gdk_property_delete (win->window,
+  gdk_property_delete (gtk_widget_get_window (win),
                        gdk_atom_intern_static_string ("WM_TRANSIENT_FOR"));
   return FALSE;
 }
@@ -171,27 +183,44 @@ ungrab_keyboard (GtkWidget *win, GdkEvent *event, gpointer data)
 static int
 delete_event (GtkWidget *widget, GdkEvent *event, gpointer data)
 {
+  (void)widget;
+  (void)event;
+  (void)data;
+
   pinentry->close_button = 1;
   gtk_main_quit ();
   return TRUE;
 }
 
 
+/* A button was clicked.  DATA indicates which button was clicked
+   (i.e., the appropriate action) and is either CONFIRM_CANCEL,
+   CONFIRM_OK or CONFIRM_NOTOK.  */
 static void
 button_clicked (GtkWidget *widget, gpointer data)
 {
+  (void)widget;
+
+  if (confirm_mode)
+    {
+      confirm_value = (confirm_value_t) data;
+      gtk_main_quit ();
+
+      return;
+    }
+
   if (data)
     {
       const char *s, *s2;
 
       /* Okay button or enter used in text field.  */
-      s = gtk_secure_entry_get_text (GTK_SECURE_ENTRY (entry));
+      s = gtk_entry_get_text (GTK_ENTRY (entry));
       if (!s)
 	s = "";
 
       if (pinentry->repeat_passphrase && repeat_entry)
         {
-          s2 = gtk_secure_entry_get_text (GTK_SECURE_ENTRY (repeat_entry));
+          s2 = gtk_entry_get_text (GTK_ENTRY (repeat_entry));
           if (!s2)
             s2 = "";
           if (strcmp (s, s2))
@@ -216,17 +245,12 @@ button_clicked (GtkWidget *widget, gpointer data)
 
 
 static void
-enter_callback (GtkWidget *widget, GtkWidget *anentry)
+enter_callback (GtkWidget *widget, GtkWidget *next_widget)
 {
-  button_clicked (widget, (gpointer) CONFIRM_OK);
-}
-
-
-static void
-confirm_button_clicked (GtkWidget *widget, gpointer data)
-{
-  confirm_value = (confirm_value_t) data;
-  gtk_main_quit ();
+  if (next_widget)
+    gtk_widget_grab_focus (next_widget);
+  else
+    button_clicked (widget, (gpointer) CONFIRM_OK);
 }
 
 
@@ -234,14 +258,12 @@ static void
 cancel_callback (GtkAccelGroup *acc, GObject *accelerable,
                  guint keyval, GdkModifierType modifier, gpointer data)
 {
-  int confirm_mode = !!data;
+  (void)acc;
+  (void)keyval;
+  (void)modifier;
+  (void)data;
 
-  if (confirm_mode)
-    confirm_button_clicked (GTK_WIDGET (accelerable),
-                            (gpointer)CONFIRM_CANCEL);
-  else
-    button_clicked (GTK_WIDGET (accelerable),
-                    (gpointer)CONFIRM_CANCEL);
+  button_clicked (GTK_WIDGET (accelerable), (gpointer)CONFIRM_CANCEL);
 }
 
 
@@ -287,14 +309,14 @@ changed_text_handler (GtkWidget *widget)
 
   if (pinentry->repeat_passphrase && repeat_entry)
     {
-      gtk_secure_entry_set_text (GTK_SECURE_ENTRY (repeat_entry), "");
+      gtk_entry_set_text (GTK_ENTRY (repeat_entry), "");
       gtk_label_set_text (GTK_LABEL (error_label), "");
     }
 
   if (!qualitybar || !pinentry->quality_bar)
     return;
 
-  s = gtk_secure_entry_get_text (GTK_SECURE_ENTRY (widget));
+  s = gtk_entry_get_text (GTK_ENTRY (widget));
   if (!s)
     s = "";
   length = strlen (s);
@@ -348,16 +370,19 @@ timeout_cb (gpointer data)
 
 
 static GtkWidget *
-create_window (pinentry_t ctx, int confirm_mode)
+create_window (pinentry_t ctx)
 {
   GtkWidget *w;
   GtkWidget *win, *box;
   GtkWidget *wvbox, *chbox, *bbox;
   GtkAccelGroup *acc;
-  GClosure *acc_cl;
   gchar *msg;
 
+  repeat_entry = NULL;
+
+#if !GTK_CHECK_VERSION (2, 12, 0)
   tooltips = gtk_tooltips_new ();
+#endif
 
   /* FIXME: check the grabbing code against the one we used with the
      old gpg-agent */
@@ -471,15 +496,13 @@ create_window (pinentry_t ctx, int confirm_mode)
 			    GTK_FILL, GTK_FILL, 4, 0);
 	}
 
-      entry = gtk_secure_entry_new ();
+      entry = gtk_entry_new ();
+      gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
       gtk_widget_set_size_request (entry, 200, -1);
-      g_signal_connect (G_OBJECT (entry), "activate",
-			G_CALLBACK (enter_callback), entry);
       g_signal_connect (G_OBJECT (entry), "changed",
                         G_CALLBACK (changed_text_handler), entry);
       gtk_table_attach (GTK_TABLE (table), entry, 1, 2, nrow, nrow+1,
                         GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 0, 0);
-      gtk_widget_grab_focus (entry);
       gtk_widget_show (entry);
       nrow++;
 
@@ -492,14 +515,19 @@ create_window (pinentry_t ctx, int confirm_mode)
 	  gtk_table_attach (GTK_TABLE (table), w, 0, 1, nrow, nrow+1,
 			    GTK_FILL, GTK_FILL, 4, 0);
 	  qualitybar = gtk_progress_bar_new();
-	  gtk_widget_add_events (qualitybar,
-				 GDK_ENTER_NOTIFY_MASK | GDK_LEAVE_NOTIFY_MASK);
 	  gtk_progress_bar_set_text (GTK_PROGRESS_BAR (qualitybar),
 				     QUALITYBAR_EMPTY_TEXT);
 	  gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (qualitybar), 0.0);
           if (pinentry->quality_bar_tt)
-            gtk_tooltips_set_tip (GTK_TOOLTIPS (tooltips), qualitybar,
-                                  pinentry->quality_bar_tt, "");
+	    {
+#if !GTK_CHECK_VERSION (2, 12, 0)
+	      gtk_tooltips_set_tip (GTK_TOOLTIPS (tooltips), qualitybar,
+				    pinentry->quality_bar_tt, "");
+#else
+	      gtk_widget_set_tooltip_text (qualitybar,
+					   pinentry->quality_bar_tt);
+#endif
+	    }
 	  gtk_table_attach (GTK_TABLE (table), qualitybar, 1, 2, nrow, nrow+1,
 	  		    GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 0, 0);
           nrow++;
@@ -515,16 +543,23 @@ create_window (pinentry_t ctx, int confirm_mode)
 	  gtk_table_attach (GTK_TABLE (table), w, 0, 1, nrow, nrow+1,
 			    GTK_FILL, GTK_FILL, 4, 0);
 
-          repeat_entry = gtk_secure_entry_new ();
+          repeat_entry = gtk_entry_new ();
+	  gtk_entry_set_visibility (GTK_ENTRY (repeat_entry), FALSE);
           gtk_widget_set_size_request (repeat_entry, 200, -1);
-          g_signal_connect (G_OBJECT (entry), "activate",
-                            G_CALLBACK (enter_callback), repeat_entry);
           gtk_table_attach (GTK_TABLE (table), repeat_entry, 1, 2, nrow, nrow+1,
                             GTK_EXPAND|GTK_FILL, GTK_EXPAND|GTK_FILL, 0, 0);
-          gtk_widget_grab_focus (entry);
-          gtk_widget_show (entry);
+          gtk_widget_show (repeat_entry);
           nrow++;
+
+	  g_signal_connect (G_OBJECT (repeat_entry), "activate",
+			    G_CALLBACK (enter_callback), NULL);
         }
+
+      /* When the user presses enter in the entry widget, the widget
+	 is activated.  If we have a repeat entry, send the focus to
+	 it.  Otherwise, activate the "Ok" button.  */
+      g_signal_connect (G_OBJECT (entry), "activate",
+			G_CALLBACK (enter_callback), repeat_entry);
     }
 
   bbox = gtk_hbutton_box_new ();
@@ -579,18 +614,15 @@ create_window (pinentry_t ctx, int confirm_mode)
             gtk_button_set_image (GTK_BUTTON (w), image);
         }
       else
-          w = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
+	w = gtk_button_new_from_stock (GTK_STOCK_CANCEL);
       gtk_container_add (GTK_CONTAINER (bbox), w);
       g_signal_connect (G_OBJECT (w), "clicked",
-                        G_CALLBACK (confirm_mode ? confirm_button_clicked
-                                    : button_clicked),
+                        G_CALLBACK (button_clicked),
 			(gpointer) CONFIRM_CANCEL);
 
-      acc_cl = g_cclosure_new (G_CALLBACK (cancel_callback),
-			       (confirm_mode? "":NULL), NULL);
-      gtk_accel_group_connect (acc, GDK_KEY_Escape, 0, 0, acc_cl);
-
-      GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
+      gtk_accel_group_connect (acc, GDK_KEY_Escape, 0, 0,
+			       g_cclosure_new (G_CALLBACK (cancel_callback),
+					       NULL, NULL));
     }
 
   if (confirm_mode && !pinentry->one_button && pinentry->notok)
@@ -601,9 +633,8 @@ create_window (pinentry_t ctx, int confirm_mode)
 
       gtk_container_add (GTK_CONTAINER (bbox), w);
       g_signal_connect (G_OBJECT (w), "clicked",
-                        G_CALLBACK (confirm_button_clicked),
+                        G_CALLBACK (button_clicked),
 			(gpointer) CONFIRM_NOTOK);
-      GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
     }
 
   if (pinentry->ok)
@@ -629,22 +660,14 @@ create_window (pinentry_t ctx, int confirm_mode)
   gtk_container_add (GTK_CONTAINER(bbox), w);
   if (!confirm_mode)
     {
-      g_signal_connect (G_OBJECT (w), "clicked",
-			G_CALLBACK (button_clicked), "ok");
       GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
       gtk_widget_grab_default (w);
-      g_signal_connect_object (G_OBJECT (entry), "focus_in_event",
-				G_CALLBACK (gtk_widget_grab_default),
-			       G_OBJECT (w), 0);
-    }
-  else
-    {
-      g_signal_connect (G_OBJECT (w), "clicked",
-			G_CALLBACK(confirm_button_clicked),
-			(gpointer) CONFIRM_OK);
-      GTK_WIDGET_SET_FLAGS (w, GTK_CAN_DEFAULT);
     }
 
+  g_signal_connect (G_OBJECT (w), "clicked",
+		    G_CALLBACK(button_clicked),
+		    (gpointer) CONFIRM_OK);
+
   gtk_window_set_position (GTK_WINDOW (win), GTK_WIN_POS_CENTER);
   gtk_window_set_keep_above (GTK_WINDOW (win), TRUE);
   gtk_widget_show_all (win);
@@ -667,7 +690,8 @@ gtk_cmd_handler (pinentry_t pe)
   pinentry = pe;
   confirm_value = CONFIRM_CANCEL;
   passphrase_ok = 0;
-  w = create_window (pe, want_pass ? 0 : 1);
+  confirm_mode = want_pass ? 0 : 1;
+  w = create_window (pe);
   gtk_main ();
   gtk_widget_destroy (w);
   while (gtk_events_pending ())
@@ -702,18 +726,6 @@ pinentry_cmd_handler_t pinentry_cmd_handler = gtk_cmd_handler;
 int
 main (int argc, char *argv[])
 {
-  static GMemVTable secure_mem =
-    {
-      secentry_malloc,
-      secentry_realloc,
-      secentry_free,
-      NULL,
-      NULL,
-      NULL
-    };
-
-  g_mem_set_vtable (&secure_mem);
-
   pinentry_init (PGMNAME);
 
 #ifdef FALLBACK_CURSES
